﻿#region File Description
//-----------------------------------------------------------------------------
// InstancedModel.cs
//
// Microsoft XNA Community Game Platform
// Copyright (C) Microsoft Corporation. All rights reserved.
// このソースコードはクリエータークラブオンラインのMeshInstancingの
// コードのコメントを翻訳、変更したもの
// http://creators.xna.com/en-US/sample/meshinstancing
//-----------------------------------------------------------------------------
#endregion

#region Using ステートメント

using System;
using System.Collections.Generic;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

#endregion

namespace DemoTypes
{
    /// <summary>
    /// 使用するインスタンス手法の列挙
    /// </summary>
    public enum InstancingTechnique
    {
        // ハードウェアインスタンス
        HardwareInstancing,

        // シェーダーインスタンス
        ShaderInstancing,
    }

    /// <summary>
    /// インスタンス技法を使って大量のモデルを描画する為のカスタムモデルクラス
    /// クリエータークラブオンラインのサンプルをシンプルにし、
    /// ハードウェアインスタンス技法をWindows、Xbox 360の両プラットフォームで
    /// 使えるようにしたもの。
    /// </summary>
    public class InstancedModel
    {
        #region プロパティ

        /// <summary>
        /// 現在のインスタンス手法を取得する
        /// </summary>
        public InstancingTechnique InstancingTechnique { get; private set; }

        /// <summary>
        /// 任意のオブジェクトを設定できるタグの取得と設定
        /// </summary>
        public object Tag { get; set; }

        /// <summary>
        /// モデルパーツリストの取得
        /// </summary>
        public List<InstancedModelPart> ModelParts { get { return modelParts; } }

        #endregion

        #region フィールド

        // モデルパーツリストの内部変数
        // 実際の描画コードの殆どはInstancedModelPartクラスに存在する
        List<InstancedModelPart> modelParts = new List<InstancedModelPart>();

        // GraphiDevice参照
        GraphicsDevice graphicsDevice;

        #endregion

        #region 初期化

        /// <summary>
        /// XNBフォーマットからインスタンスモデルデータを読み出すコンストラクタ
        /// </summary>
        internal InstancedModel( ContentReader input )
        {
            // ContentReaderからGraphicsDeviceを取得する
            graphicsDevice = RenderHelper.GetGraphicsDevice( input );

            // モデルデータの読み込み
            int partCount = input.ReadInt32();

            for ( int i = 0; i < partCount; i++ )
            {
                modelParts.Add( new InstancedModelPart( this, input, graphicsDevice ) );
            }

            // タグに割り当てられているデータの読み込み
            Tag = input.ReadObject<object>();
        }

        #endregion

        #region インスタンス手法の選択

        /// <summary>
        /// インスタンス手法の変更
        /// </summary>
        /// <param name="technique">インスタンス手法</param>
        /// <param name="instanceVertexElements">
        /// HWインスタンスで使うインスタンス用の頂点要素</param>
        /// <returns>指定した手法が使えるならtrueを返し、
        /// サポートされていなければfalseを返す</returns>
        public bool SetInstancingTechnique(
                    InstancingTechnique technique,
                    VertexElement[] instanceVertexElements )
        {
            if ( !IsTechniqueSupported( technique ) )
                return false;

            InstancingTechnique = technique;

            foreach ( InstancedModelPart modelPart in modelParts )
            {
                modelPart.Initialize( technique, instanceVertexElements );
            }

            return true;
        }

        /// <summary>
        /// GPUが指定されたインスタンス技法をサポートしているか調べる
        /// </summary>
        public bool IsTechniqueSupported( InstancingTechnique technique )
        {
            // HWインスタンスを使うにはシェーダーモデル3.0が必要
            if ( technique == InstancingTechnique.HardwareInstancing )
            {
                GraphicsDeviceCapabilities caps =
                                            graphicsDevice.GraphicsDeviceCapabilities;
                return caps.PixelShaderVersion.Major >= 3;
            }

            // HWインスタンス以外の手法はXNAフレームワークが動作するどのGPUでも使用可能
            return true;
        }

        #endregion
    }
}
