﻿#region File Description
//-----------------------------------------------------------------------------
// ShaderInstancePart.cs
//
// Microsoft XNA Community Game Platform
// Copyright (C) Microsoft Corporation. All rights reserved.
// このソースコードはクリエータークラブオンラインのMeshInstancingの
// コードのコメントを翻訳、変更したもの
// http://creators.xna.com/en-US/sample/meshinstancing
//-----------------------------------------------------------------------------
#endregion

#region Using ステートメント

using System;

using Microsoft.Xna.Framework.Graphics;

#endregion

namespace DemoTypes
{
    /// <summary>
    /// シェーダーインスタンス用の初期化コードとパラメーター
    /// シェーダーインスタンスは定数レジスタにインスタンス情報を格納することで、
    /// インスタンスモデル描画を実現している。
    /// 
    /// 定数レジスタを使用しているので、SM 1.1では96個、SM 2.0では256個以上の
    /// 定数レジスタの上限数に気をつける必要がある。
    /// 
    /// オリジナルのサンプルでは4x4の行列を直接渡していたので最大60個の
    /// インスタンスしか描画できないが、このデモでは座標(Vector3)、スケール(float)、
    /// そして任意の軸の回転情報(Vector4)をインスタンスの情報としているので、
    /// インスタンスの最大数が100個になり、余った定数レジスタはBasicEffect同じ
    /// 光源処理をするのに使用している。
    /// 
    /// また、シェーダーインスタンスでは頂点バッファとインデックスバッファのサイズが
    /// オリジナルのサイズよりインスタンスの数の分だけ倍増することに注意すること
    /// </summary>
    public partial class InstancedModelPart
    {
        #region 定数宣言

        /// <summary>
        /// シェーダーインスタンス技法を使った場合の最大インスタンス数 
        /// </summary>
        /// <remarks>この定数を変更するときにはInstancedModel.fx内の
        /// MAX_SHADER_INSTANCE_COUNTを変更することを忘れないこと</remarks>
        const int MaxShaderInstanceCount = 100;

        #endregion

        #region フィールド

        //シェーダーインスタンス用の頂点バッファとインデックスバッファ
        VertexBuffer shaderInstanceVB;
        IndexBuffer shaderInstanceIB;
        int shaderInstanceVertexStride;
        VertexDeclaration shaderInstanceVDecl;

        bool shaderInstanceInitialized;

        #endregion

        #region 初期化

        /// <summary>
        /// シェーダーインスタンス用にジオメトリを初期化する
        /// </summary>
        void InitializeShaderInstancing( bool disposeOriginalResources )
        {
            // ひとつのバッチにどれだけのインスタンス数が描画できるか計算する
            int indexOverflowLimit = ushort.MaxValue / vertexCount;

            MaxInstances = Math.Min( indexOverflowLimit, MaxShaderInstanceCount );

            // 初期化が一回だけ必要なものを初期化する
            if ( !shaderInstanceInitialized )
            {
                ReplicateVertexData();
                ReplicateIndexData();

                // シェーダーインスタンスを使う場合、各頂点にインスタンス番号を格納する
                // 必要がある。ここでは、テクスチャ座標のチャンネル1にその情報を
                // 追加している。
                int instanceIndexOffset = vertexStride;
                byte usageIndex = 1;
                short stream = 0;

                VertexElement[] extraElements =
                {
                    new VertexElement(stream, (short)instanceIndexOffset,
                                      VertexElementFormat.Single,
                                      VertexElementMethod.Default,
                                      VertexElementUsage.TextureCoordinate, usageIndex)
                };

                shaderInstanceVertexStride = vertexStride + sizeof( float );
                shaderInstanceVDecl = RenderHelper.ExtendVertexDeclaration(
                                                    vertexDeclaration.GraphicsDevice,
                                                    originalVertexDeclaration,
                                                    extraElements );

                shaderInstanceInitialized = true;
            }

            // オリジナルの頂点バッファとインデックスバッファを破棄する
            if ( disposeOriginalResources )
            {
                if ( vertexBuffer != null )
                {
                    vertexBuffer.Dispose();
                    vertexBuffer = null;
                }

                if ( indexBuffer != null )
                {
                    indexBuffer.Dispose();
                    indexBuffer = null;
                }
            }
        }

        /// <summary>
        /// 頂点バッファの複製
        /// 頂点バッファをインスタンスの分だけコピーし、それぞれの頂点に
        /// インスタンス番号の情報を追加する
        /// </summary>
        void ReplicateVertexData()
        {
            // オリジナルの頂点バッファのデータを読み込んでから破棄する
            byte[] oldVertexData = new byte[vertexCount * vertexStride];

            vertexBuffer.GetData( oldVertexData );

            // ストライドサイズを追加するチャンネル分調整する
            shaderInstanceVertexStride = vertexStride + sizeof( float );

            int stride = shaderInstanceVertexStride;

            // 複製した頂点データを入れるためのメモリを確保する
            byte[] newVertexData = new byte[vertexCount * stride * MaxInstances];

            int outputPosition = 0;

            // インスタンスの分だけオリジナルの頂点データを複製する
            for ( int instanceIndex = 0; instanceIndex < MaxInstances; instanceIndex++ )
            {
                int sourcePosition = 0;

                // インスタンスインデックスをfloatからビットデータに変換
                byte[] instanceIndexBits = BitConverter.GetBytes( (float)instanceIndex );

                for ( int i = 0; i < vertexCount; i++ )
                {
                    // 既存の頂点データをコピーする
                    Array.Copy( oldVertexData, sourcePosition,
                               newVertexData, outputPosition, vertexStride );

                    outputPosition += vertexStride;
                    sourcePosition += vertexStride;

                    // 新しいインデックスチャンネル値を設定する
                    instanceIndexBits.CopyTo( newVertexData, outputPosition );

                    outputPosition += instanceIndexBits.Length;
                }
            }

            // 新しい頂点バッファを作り、複製したデータを設定する
            shaderInstanceVB = new VertexBuffer( vertexBuffer.GraphicsDevice,
                                            newVertexData.Length, BufferUsage.None );

            shaderInstanceVB.SetData( newVertexData );
        }


        /// <summary>
        /// インデックスバッファの複製
        /// 最大インスタンス分だけ複製し、インスタンス毎のインデックスを設定する
        /// </summary>
        void ReplicateIndexData()
        {
            // オリジナルのインデックスバッファのデータを読み込んでから破棄する
            ushort[] oldIndices = new ushort[indexCount];

            indexBuffer.GetData( oldIndices );

            // 複製するインデックス用のメモリを確保
            ushort[] newIndices = new ushort[indexCount * MaxInstances];

            int outputPosition = 0;

            // 各インスタンスにオリジナルのインデックスを複製する
            for ( int instanceIndex = 0; instanceIndex < MaxInstances; instanceIndex++ )
            {
                int instanceOffset = instanceIndex * vertexCount;

                for ( int i = 0; i < indexCount; i++ )
                {
                    newIndices[outputPosition] = (ushort)( oldIndices[i] +
                                                          instanceOffset );

                    outputPosition++;
                }
            }

            // 新しいインデックスバッファを生成し、複製したデータを設定する
            shaderInstanceIB = new IndexBuffer( indexBuffer.GraphicsDevice,
                                          sizeof( ushort ) * newIndices.Length,
                                          BufferUsage.None,
                                          IndexElementSize.SixteenBits );

            shaderInstanceIB.SetData( newIndices );
        }

        #endregion

    }
}
