﻿//-----------------------------------------------------------------------------
// InstancedModel.fx
//
// Microsoft XNA Community Game Platform
// Copyright (C) Microsoft Corporation. All rights reserved.
// このソースコードはクリエータークラブオンラインのMeshInstancingの
// コードのコメントを翻訳、変更したもの
// http://creators.xna.com/en-US/sample/meshinstancing
//=============================================================================
#include "Basic.fx"

// シェーダーインスタンス時に描画できる最大インスタンス数は頂点シェーダーの
// 定数レジスタに左右される。シェーダーモデル2.0では256の定数レジスタが使えるので
// ここでは100個のインスタンスを描画できるようにしある。
// 他にも定数レジスタが必要な場合や、96個の定数レジスタ数制限がある
// シェーダーモデル 1.1ではMAX_SHADER_INSTANCE_COUNTの値を減らす必要がある。
// この値を減らした場合、ShaderInstancePart.csのMaxShaderInstanceCount定数を
// 変えることを忘れないこと。

#define MAX_SHADER_INSTANCE_COUNT 100

// シェーダーインスタンス用のパラメーター格納用配列
float4 ShaderInstanceParams0[MAX_SHADER_INSTANCE_COUNT];
float4 ShaderInstanceParams1[MAX_SHADER_INSTANCE_COUNT];

struct VertexShaderInput
{
    float4 Position : POSITION0;
    float3 Normal	: NORMAL0;
    float2 TexCoord : TEXCOORD0;
};

struct VertexShaderOutput
{
    float4 Position : POSITION0;
    float4 Diffuse	: COLOR0;
    float4 Specular	: COLOR1;
    float2 TexCoord : TEXCOORD0;
};

//-----------------------------------------------------------------------------
// 任意の回転軸で回転するクォータニオンの生成
//=============================================================================
float4 QuaternionFromAxisAngle( float3 axis, float angle)
{
	float halfAngle = angle * 0.5f;
	return float4( axis * sin(halfAngle), cos(halfAngle));
}

//-----------------------------------------------------------------------------
// クォータニオンによる頂点変換
//=============================================================================
float3 TransformByQuaternion( float3 position, float4 quaternion)
{
	float4 Q = quaternion;
	float3 v = position;
	
	return 
		( 2.0f * Q.w * Q.w - 1.0f ) * v +
		( 2.0f * dot( v, Q.xyz ) * Q.xyz ) +
		( 2.0f * Q.w * cross( Q.xyz, v ) );
}

//-----------------------------------------------------------------------------
// 各頂点シェーダーから呼ばれる共通の頂点処理
//=============================================================================
VertexShaderOutput VertexShaderCommon( VertexShaderInput input, float4 instanceParam0,
																 float4 instanceParam1)
{
    VertexShaderOutput output;
    
    // 頂点変換
    float3 pos_os = instanceParam0.xyz;
    float scale = instanceParam0.w;
    float3 rotationAxis = instanceParam1.xyz;
    float rotation = instanceParam1.w;
    
    float4 Q = QuaternionFromAxisAngle( rotationAxis, rotation );
    float4 pos_ws = float4( pos_os +
								TransformByQuaternion( input.Position * scale, Q ), 1 );
 
    float4 pos_vs = mul(pos_ws, View);
    output.Position = mul(pos_vs, Projection);

    // 光源処理
    float3 N = TransformByQuaternion( input.Normal, Q );
	float3 posToEye = EyePosition - pos_ws;
	float3 E = normalize(posToEye);
	ColorPair lightResult = ComputeLights(E, N);
    
    output.Diffuse = float4(lightResult.Diffuse, 1);
    output.Specular = float4( lightResult.Specular, 1 );

    // 入力テクスチャ座標をそのまま出力
    output.TexCoord = input.TexCoord;

    return output;
}

//-----------------------------------------------------------------------------
// シェーダーインスタンスの頂点シェーダー
// インスタンスインデックスは頂点データ内に格納されている
//=============================================================================
VertexShaderOutput ShaderInstancingVS(VertexShaderInput input,
                                                float instanceIndex : TEXCOORD1)
{
    return VertexShaderCommon(input,
		ShaderInstanceParams0[instanceIndex], ShaderInstanceParams1[instanceIndex]);
}

#ifdef XBOX360

//-----------------------------------------------------------------------------
// HWインスタンスの頂点シェーダー(Xbox 360版)
// Xbox 360にはHWインスタンス機能は内がvfetchを使うことでWindowsに近い実装
// ここでは頂点処理は行わず、vfetchを使って頂点データとインスタンス情報を
// 取得してWindows/Xbox 360共通の頂点シェーダーを呼んでいる
//=============================================================================

// 描画対象モデルの頂点数
int VertexCount;

VertexShaderOutput HardwareInstancingVS(int index : INDEX)
{
	// 現在のインデックス値から頂点インデックスとインスタンスインデックスを計算する
    int vertexIndex = ( index + 0.5 ) % VertexCount;
    int instanceIndex = ( index + 0.5 ) / VertexCount;

    float4 position;
    float4 normal;
    float4 texCoord;
    float4 instanceParam0;
    float4 instanceParam1;

	// vfetchはアセンブリ命令なのでasmブロックを使う必要がある
    asm
    {
		// 頂点データのフェッチ
        vfetch position,		vertexIndex, position0
        vfetch normal,			vertexIndex, normal0
        vfetch texCoord,		vertexIndex, texcoord0
        
		// インスタンスデータのフェッチ
        vfetch instanceParam0,	instanceIndex, texcoord1
        vfetch instanceParam1,	instanceIndex, texcoord2
    };

	// 共通頂点シェーダーを呼び出す
    VertexShaderInput input;

    input.Position = position;
    input.Normal = normal;
    input.TexCoord = texCoord;

    return VertexShaderCommon(input, instanceParam0, instanceParam1 );
}

#else

//-----------------------------------------------------------------------------
// HWインスタンスの頂点シェーダー(Windows版)
// Windows版でシェーダーモデル3.0以降のカードではHWインスタンスが使える
// インスタンス毎の情報は２番目の頂点ストリームから読み込まれる
//=============================================================================
VertexShaderOutput HardwareInstancingVS(VertexShaderInput input,
                                                float4 instanceParam0 : TEXCOORD1,
                                                float4 instanceParam1 : TEXCOORD2 )
{
    return VertexShaderCommon(input, instanceParam0, instanceParam1 );
}

#endif

//-----------------------------------------------------------------------------
// 共通のピクセルシェーダー
//=============================================================================
float4 PS( VertexShaderOutput pin ) : COLOR0
{
	return tex2D( TextureSampler, pin.TexCoord ) * 
					pin.Diffuse + float4(pin.Specular.rgb, 0);
}

//-----------------------------------------------------------------------------
// 各テクニックの宣言
//=============================================================================
// シェーダーインスタンス
technique ShaderInstancing
{
    pass Pass1
    {
        VertexShader = compile vs_2_0 ShaderInstancingVS();
        PixelShader = compile ps_1_1 PS();
    }
}

// HWインスタンス
// Xbox 360またはSM 3.0以上のビデオカードが必要
technique HardwareInstancing
{
    pass Pass1
    {
        VertexShader = compile vs_3_0 HardwareInstancingVS();
        PixelShader = compile ps_3_0 PS();
    }
}
