﻿using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

using System.Runtime.InteropServices;   // StructLayoutアトリビュートの為

namespace ShaderInstancingTest02
{
    // class GameObject
    // {

    /// <summary>
    /// ゲームオブジェクト構造体
    /// </summary>
    [StructLayout( LayoutKind.Sequential )]
    public struct GameObject
    {
        // CPUで更新し、GPUで読み込まれる変数
        public Vector3  Position;       // 座標
        public float    Scale;          // スケール
        public Vector3  RotateAxis;     // 回転軸
        public float    Rotation;       // 回転

        // CPUのみで使用する変数
        public Vector3  Velocity;       // 速度
        public float    RotationSpeed;  // 回転速度
        public float    LifeTime;       // 寿命

        /// <summary>
        /// ゲームオブジェクトの初期化
        /// </summary>
        public void Initialize( Random rnd )
        {
            // 出現位置、移動速度などをランダムに生成する
            float w = Game1.Sandbox.Max.X - Game1.Sandbox.Min.X;
            float h = Game1.Sandbox.Max.Y - Game1.Sandbox.Min.Y;

            Position = new Vector3(
                Game1.Sandbox.Min.X + w * (float)rnd.NextDouble(),
                Game1.Sandbox.Min.Y + h * (float)rnd.NextDouble(),
                (float)rnd.NextDouble() * -50.0f );

            float rad = (float)rnd.NextDouble() * MathHelper.TwoPi;
            Velocity = new Vector3( (float)Math.Cos( rad ),
                                    (float)Math.Sin( rad ), 0 ) * 3.0f;

            float theta = (float)rnd.NextDouble() * MathHelper.TwoPi;
            float phi = (float)rnd.NextDouble() * MathHelper.TwoPi;
            RotateAxis = Vector3.Transform( Vector3.UnitX,
                                Matrix.CreateFromYawPitchRoll( theta, 0, phi ) );
            Rotation = (float)rnd.NextDouble() * MathHelper.TwoPi;
            RotationSpeed = (float)rnd.NextDouble() * MathHelper.TwoPi;

            Scale = 0.5f + (float)rnd.NextDouble();
            LifeTime = 20.0f + 10.0f * (float)rnd.NextDouble();
        }

        /// <summary>
        /// ゲームオブジェクトの更新
        /// </summary>
        /// <param name="dt">更新時間</param>
        /// <returns>寿命が尽きたらfalseを返す</returns>
        public bool Update( float dt )
        {
            // 座標と回転の更新
            Position.X += Velocity.X * dt;
            Position.Y += Velocity.Y * dt;
            Position.Z += Velocity.Z * dt;
            Rotation = FastMath.NormalizeRotation( Rotation + RotationSpeed * dt );

            // 移動範囲の端に来たら跳ね返る処理
            if ((Position.X < Game1.Sandbox.Min.X && Velocity.X < 0) ||
                    (Position.X > Game1.Sandbox.Max.X && Velocity.X > 0))
            {
                Velocity.X = -Velocity.X;
            }

            if ((Position.Y < Game1.Sandbox.Min.Y && Velocity.Y < 0) ||
                (Position.Y > Game1.Sandbox.Max.Y && Velocity.Y > 0))
            {
                Velocity.Y = -Velocity.Y;
            }

            // 寿命の更新
            LifeTime -= dt;

            return LifeTime > 0.0f;
        }

        /// <summary>
        /// 頂点要素
        /// </summary>
        static public VertexElement[] VertexElements = {
            // Position (8バイト)
            new VertexElement(0, 0, VertexElementFormat.Vector2,
                                    VertexElementMethod.Default,
                                    VertexElementUsage.Position, 0),
            // Color  (4バイト)
            new VertexElement(0, 8, VertexElementFormat.Color,
                                    VertexElementMethod.Default,
                                    VertexElementUsage.Color, 0),
            // Scale, Rotation, Depth (12バイト)
            new VertexElement(0,12, VertexElementFormat.Vector3,
                                    VertexElementMethod.Default,
                                    VertexElementUsage.TextureCoordinate, 0),
        };

        /// <summary>
        /// この構造体のサイズを返す
        /// </summary>
        public static int SizeInBytes
        {
            get { return Marshal.SizeOf( typeof( GameObject ) ); }
        }
    }
}
